/*	com.c

	AFAPI communication function definitions for PAPERS_JR.

	July 1996 by H. Dietz
*/


#include "afapi.h"
#include "inline.h"

/*	p_bcastPut/Get() for 1u, 8u, 8, and bar...
*/

inline void
p_bcastPut1u(register p_1u datum)
{
	p__put(datum != 0);
}


inline p_1u
p_bcastGet1u(void)
{
	return(p__get());
}


inline void
p_bcastPut8u(register p_8u datum)
{
	p__put(datum >> 4);
	p__put(datum);
}

inline p_8u
p_bcastGet8u(void)
{
	register p_8u high = p__get();
	return(p__get() | (high << 4));
}

inline void
p_bcastPut8(register p_8 datum)
{
	p__put(datum >> 4);
	p__put(datum);
}

inline p_8
p_bcastGet8(void)
{
	register p_8 high = p__get();
	return(p__get() | (high << 4));
}

void
p_bcastPutbar(register p_bar datum)
{
	/* Hmm...  bar is only 2 bits, which is < 4... */
	p__put(datum);
}

p_bar
p_bcastGetbar(void)
{
	/* Hmm...  bar is only 2 bits, which is < 4... */
	return(p__get());
}

/*	p_bcastPut/Get() for 16u, 16, 32, 32u, 64, 64u...

	Each of these routines uses a nybble-prefix tag to efficiently
	take advantage of the fact that the most significant bits of
	most 16 bit or larger integers are very often all 0s or all 1s.
	Imagine dividing each integer into four segments.  The first
	tag bit indicates if the most significant quarter is 0.  The
	remaining three bits indicate if the previous quarter and the
	next quarter have the same value.  64-bit stuff is actually
	treated as two 32-bit values.
*/
	
void
p_bcastPut16u(register p_16u datum)
{
	/* Use simple compression algorithm...
	   4-bit tag says which nybbles differ
	*/
	register p_8u c0 = (datum & 0xf);
	register p_8u c1 = ((datum >> 4) & 0xf);
	register p_8u c2 = ((datum >> 8) & 0xf);
	register p_8u c3 = ((datum >> 12) & 0xf);
	register int tag = (((c0 == c1) ? 0 : 1) |
			    ((c1 == c2) ? 0 : 2) |
			    ((c2 == c3) ? 0 : 4) |
			    ((c3 == 0) ? 0 : 8));
	p__put(tag);
	if (tag & 8) p__put(c3);
	if (tag & 4) p__put(c2);
	if (tag & 2) p__put(c1);
	if (tag & 1) p__put(c0);
}

p_16u
p_bcastGet16u(void)
{
	/* Use simple compression algorithm...
	   4-bit tag says which bytes differ
	*/
	register p_16u datum = 0;
	register p_8u c;
	register int tag = p__get();

	datum = (c = ((tag & 8) ? p__get() : 0));
	if (tag & 4) c = p__get();
	datum = ((datum << 4) | c);
	if (tag & 2) c = p__get();
	datum = ((datum << 4) | c);
	if (tag & 1) c = p__get();
	return((datum << 4) | c);
}


void
p_bcastPut16(register p_16 datum)
{
	p_bcastPut16u((p_16u) datum);
}


p_16
p_bcastGet16(void)
{
	return((p_16) p_bcastGet16u());
}


void
p_bcastPut32u(register p_32u datum)
{
	/* Use simple compression algorithm...
	   4-bit tag says which bytes differ
	*/
	register p_8u c0 = (datum & 0xff);
	register p_8u c1 = ((datum >> 8) & 0xff);
	register p_8u c2 = ((datum >> 16) & 0xff);
	register p_8u c3 = ((datum >> 24) & 0xff);
	register int tag = (((c0 == c1) ? 0 : 1) |
			    ((c1 == c2) ? 0 : 2) |
			    ((c2 == c3) ? 0 : 4) |
			    ((c3 == 0) ? 0 : 8));
	p__put(tag);
	if (tag & 8) p_bcastPut8u(c3);
	if (tag & 4) p_bcastPut8u(c2);
	if (tag & 2) p_bcastPut8u(c1);
	if (tag & 1) p_bcastPut8u(c0);
}


p_32u
p_bcastGet32u(void)
{
	/* Use simple compression algorithm...
	   4-bit tag says which bytes differ
	*/
	register p_32u datum = 0;
	register p_8u c;
	register int tag = p__get();

	datum = (c = ((tag & 8) ? p_bcastGet8u() : 0));
	if (tag & 4) c = p_bcastGet8u();
	datum = ((datum << 8) | c);
	if (tag & 2) c = p_bcastGet8u();
	datum = ((datum << 8) | c);
	if (tag & 1) c = p_bcastGet8u();
	return((datum << 8) | c);
}


void
p_bcastPut32(register p_32 datum)
{
	p_bcastPut32u((p_32u) datum);
}


p_32
p_bcastGet32(void)
{
	return((p_32) p_bcastGet32u());
}


void
p_bcastPut64u(register p_64u datum)
{
	p_bcastPut32u((p_32u) (datum >> 32));
	p_bcastPut32u((p_32u) datum);
}

p_64u
p_bcastGet64u(void)
{
	register p_64u result = p_bcastGet32u();
	return((result << 32) | p_bcastGet32u());
}

void
p_bcastPut64(register p_64 datum)
{
	p_bcastPut64u((p_64u) datum);
}

p_64
p_bcastGet64()
{
	return((p_64) p_bcastGet64u());
}


void
p_bcastPutptr(register p_ptr datum)
{
	p_bcastPut32u((p_32u) datum);
}


p_ptr
p_bcastGetptr(void)
{
	return((p_ptr) p_bcastGet32u());
}

/*	p_bcastPutz() and p_bcastGetz()...

	We have already developed a very fancy hash-indexed, cached,
	run-length encoded, compression scheme...  but this does not
	use it.  We will distribute that version when we have more
	information about the type of data these routines are used to
	transmit, so that we can better tune the hashing, etc.

	For now, these routines simply use a run-length encoding
	similar to the compression used to send p_32u objects.
*/

inline void
p_bcastPutz(register void *source,
register int bytes)
{
	register p_8u prev = 0;

	while (bytes >= 4) {
		/* Use simple compression algorithm...
		   4-bit tag says which bytes differ
		*/
		register p_8u c0;
		register p_8u c1;
		register p_8u c2;
		register p_8u c3;
		register int tag;

		c0 = *((p_8u *) source);
		++source;
		c1 = *((p_8u *) source);
		++source;
		c2 = *((p_8u *) source);
		++source;
		c3 = *((p_8u *) source);
		++source;
		tag = (((c0 == prev) ? 0 : 1) |
		       ((c1 == c0) ? 0 : 2) |
		       ((c2 == c1) ? 0 : 4) |
		       ((c3 == c2) ? 0 : 8));
		p__put(tag);
		if (tag & 1) p_bcastPut8u(c0);
		if (tag & 2) p_bcastPut8u(c1);
		if (tag & 4) p_bcastPut8u(c2);
		if (tag & 8) p_bcastPut8u(c3);
		prev = c3;
		bytes -= 4;
	}

	while (bytes-- > 0) {
		p_bcastPut8u(*((p_8u *) source));
		++source;
	}
}


inline void
p_bcastGetz(register void *dest,
register int bytes)
{
	register p_8u prev = 0;

	while (bytes >= 4) {
		/* Use simple compression algorithm...
		   4-bit tag says which bytes differ
		*/
		register p_8u c0;
		register p_8u c1;
		register p_8u c2;
		register int tag = p__get();

		if (tag & 1) prev = p_bcastGet8u();
		c0 = prev;
		if (tag & 2) prev = p_bcastGet8u();
		c1 = prev;
		if (tag & 4) prev = p_bcastGet8u();
		c2 = prev;
		if (tag & 8) prev = p_bcastGet8u();
		*((p_8u *) dest) = c0;
		++dest;
		*((p_8u *) dest) = c1;
		++dest;
		*((p_8u *) dest) = c2;
		++dest;
		*((p_8u *) dest) = prev;
		++dest;
		bytes -= 4;
	}

	while (bytes-- > 0) {
		*((p_8u *) dest) = p_bcastGet8u();
		++dest;
	}
}

/*	p_bcastPut/Get() for 32f and 64f...

	Really can't easily compress floats; if you think about it, a
	floating point number has already undergone lossy compression.
	We'll use the compressed block transmission code.  Notice that
	this is not usable with a heterogeneous mix of floating point
	formats.
*/

static p_32f p_temp32f;
static p_64f p_temp64f;

void
p_bcastPut32f(register p_32f datum)
{
	p_temp32f = datum;
	p_bcastPutz(((void *) &p_temp32f), sizeof(p_32f));
}


p_32f
p_bcastGet32f(void)
{
	p_bcastGetz(((void *) &p_temp32f), sizeof(p_32f));
	return(p_temp32f);
}


void
p_bcastPut64f(register p_64f datum)
{
	p_temp64f = datum;
	p_bcastPutz(((void *) &p_temp64f), sizeof(p_64f));
}


p_64f
p_bcastGet64f(void)
{
	p_bcastGetz(((void *) &p_temp64f), sizeof(p_64f));
	return(p_temp64f);
}

/*	p_putGet() functions...

	The basic algorithm is to use p__swap() and then pick the
	datum that we want, either ours or the one just obtained.
*/

#define	P_DECLARE(suffix) \
inline p_##suffix \
p_putGet##suffix(register p_##suffix datum, \
register int source) \
{ \
	register int i = (2 * sizeof(p_##suffix)); \
	register int shift = 0; \
	register p_##suffix d = datum; \
	register p_##suffix d2 = 0; \
 \
	do { \
		d2 |= (p__swap(d) << shift); \
		d >>= 4; \
		shift += 4; \
	} while (--i > 0); \
 \
	return((source == IPROC) ? datum : d2); \
}

P_DECLARE(bar)
P_DECLARE(8)
P_DECLARE(8u)
P_DECLARE(16)
P_DECLARE(16u)
P_DECLARE(32)
P_DECLARE(32u)
P_DECLARE(64)
P_DECLARE(64u)

#undef	P_DECLARE


p_1u
p_putGet1u(register p_1u datum,
register int source)
{
	return(p_putGet8u((datum != 0), source));
}


p_32f
p_putGet32f(register p_32f datum,
register int source)
{
	p_32f t = datum;
	*((p_32u *) &t) = p_putGet32u((*((p_32u *) &t)), source);
	return(t);
}


p_64f
p_putGet64f(register p_64f datum,
register int source)
{
	/* The following cr*p is needed because GCC 64-bit int types
	   are *seriously* crippled, yielding a variety of errors that
	   we couldn't otherwise circumvent.  As a matter of policy,
	   we'd rather GCC get fixed than that we muck with all these
	   operations, so we've just "fixed" what seemed to be
	   critical....
	*/
	volatile p_64f t = datum;
	register p_8u *p = ((p_8u *) &t);
	*p = p_putGet8u(*p, source);
	++p;
	*p = p_putGet8u(*p, source);
	++p;
	*p = p_putGet8u(*p, source);
	++p;
	*p = p_putGet8u(*p, source);
	++p;
	*p = p_putGet8u(*p, source);
	++p;
	*p = p_putGet8u(*p, source);
	++p;
	*p = p_putGet8u(*p, source);
	++p;
	*p = p_putGet8u(*p, source);
	return(t);
}


p_ptr
p_putGetptr(register p_ptr datum,
register int source)
{
	p_ptr t = datum;
	*((p_32u *) &t) = p_putGet32u((*((p_32u *) &t)), source);
	return(t);
}

/*	p_gather() functions...

	With just 2 processors, gather is putGet....
*/

#define	P_DECLARE(suffix) \
inline void \
p_gather##suffix(register p_##suffix *dest, \
register p_##suffix datum) \
{ \
	dest[IPROC] = datum; \
	dest[IPROC ^ 1] = p_putGet##suffix(datum, (IPROC ^ 1)); \
}

P_ALPHA

#undef	P_DECLARE


/*	p_scatPut() and p_scatGet() functions...

	This is really a very simple operation to construct using the
	other routines, but it has been added to the library by
	popular demand....
*/

#define	P_DECLARE(suffix) \
p_##suffix \
p_scatPut##suffix(register p_##suffix *src) \
{ \
	register int who = 0; \
 \
	do { \
		p_bcastPut##suffix(src[who]); \
	} while (++who < NPROC); \
 \
	return(src[IPROC]); \
} \
 \
p_##suffix \
p_scatGet##suffix(void) \
{ \
	register p_##suffix t; \
	register p_##suffix result = 0; \
	register int who = 0; \
 \
	do { \
		t = p_bcastGet##suffix(); \
		if (who == IPROC) result = t; \
	} while (++who < NPROC); \
 \
	return(result); \
}

P_ALPHA

#undef	P_DECLARE
